'use strict';

var levenshtein = require('../../utils/levenshtein.js');
var helper = require('../../utils/helper.js');
var reverse = require('./variants/matching/reverse.js');
var l33t = require('./variants/matching/l33t.js');
var mergeUserInputDictionary = require('../../utils/mergeUserInputDictionary.js');

class MatchDictionary {
  constructor(options) {
    this.options = options;
    this.l33t = new l33t(options, this.defaultMatch);
    this.reverse = new reverse(options, this.defaultMatch);
  }
  match(matchOptions) {
    const matches = [...this.defaultMatch(matchOptions), ...this.reverse.match(matchOptions), ...this.l33t.match(matchOptions)];
    return helper.sorted(matches);
  }
  defaultMatch({
    password,
    userInputsOptions,
    useLevenshtein = true
  }) {
    const matches = [];
    const passwordLength = password.length;
    const passwordLower = password.toLowerCase();
    const {
      rankedDictionaries,
      rankedDictionariesMaxWordSize
    } = mergeUserInputDictionary(this.options.rankedDictionaries, this.options.rankedDictionariesMaxWordSize, userInputsOptions);
    // eslint-disable-next-line complexity,max-statements
    Object.keys(rankedDictionaries).forEach(dictionaryName => {
      const rankedDict = rankedDictionaries[dictionaryName];
      const longestDictionaryWordSize = rankedDictionariesMaxWordSize[dictionaryName];
      const searchWidth = Math.min(longestDictionaryWordSize, passwordLength);
      for (let i = 0; i < passwordLength; i += 1) {
        const searchEnd = Math.min(i + searchWidth, passwordLength);
        for (let j = i; j < searchEnd; j += 1) {
          const usedPassword = passwordLower.slice(i, +j + 1 || 9e9);
          const isInDictionary = (usedPassword in rankedDict);
          let foundLevenshteinDistance = {};
          // only use levenshtein distance on full password to minimize the performance drop
          // and because otherwise there would be to many false positives
          const isFullPassword = i === 0 && j === passwordLength - 1;
          if (this.options.useLevenshteinDistance && isFullPassword && !isInDictionary && useLevenshtein) {
            foundLevenshteinDistance = levenshtein(usedPassword, rankedDict, this.options.levenshteinThreshold);
          }
          const isLevenshteinMatch = Object.keys(foundLevenshteinDistance).length !== 0;
          if (isInDictionary || isLevenshteinMatch) {
            const usedRankPassword = isLevenshteinMatch ? foundLevenshteinDistance.levenshteinDistanceEntry : usedPassword;
            const rank = rankedDict[usedRankPassword];
            matches.push({
              pattern: 'dictionary',
              i,
              j,
              token: password.slice(i, +j + 1 || 9e9),
              matchedWord: usedPassword,
              rank,
              dictionaryName: dictionaryName,
              reversed: false,
              l33t: false,
              ...foundLevenshteinDistance
            });
          }
        }
      }
    });
    return matches;
  }
}

module.exports = MatchDictionary;
//# sourceMappingURL=matching.js.map
